@@ -23,7 +23,7 @@
  *
  * You should have received a copy of the GNU General Public License 
  * along with this program; if not, write to the Free Software 
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 /*
  * History:
@@ -126,6 +126,9 @@
 #include "../../atomic_ops.h" /* membar_depends() */
 
 
+extern int tm_failure_exec_mode;
+extern int tm_dns_reuse_rcv_socket;
+
 static int goto_on_branch = 0, branch_route = 0;
 
 void t_on_branch( unsigned int go_to )
@@ -138,7 +141,7 @@ void t_on_branch( unsigned int go_to )
 	if (!t || t==T_UNDEFINED ) {
 		goto_on_branch=go_to;
 	} else {
-		get_t()->on_branch = go_to;
+		t->on_branch = go_to;
 	}
 }
 
@@ -147,12 +150,12 @@ unsigned int get_on_branch(void)
 	return goto_on_branch;
 }
 
+void set_branch_route( unsigned int on_branch)
+{
+	branch_route = on_branch;
+}
 
 
-/* prepare_new_uac flags */
-#define UAC_DNS_FAILOVER_F 1 /**< new branch due to dns failover */
-#define UAC_SKIP_BR_DST_F  2 /**< don't set next hop as dst_uri for
-							   branch_route */
 
 
 /** prepares a new branch "buffer".
@@ -187,7 +190,9 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 									str* next_hop,
 									struct socket_info* fsocket,
 									snd_flags_t snd_flags,
-									int fproto, int flags)
+									int fproto, int flags,
+									str *instance, str *ruid,
+									str *location_ua)
 {
 	char *shbuf;
 	struct lump* add_rm_backup, *body_lumps_backup;
@@ -200,6 +205,12 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 	int dst_uri_backed_up;
 	str path_bak;
 	int free_path;
+	str instance_bak;
+	int free_instance;
+	str ruid_bak;
+	int free_ruid;
+	str ua_bak;
+	int free_ua;
 	int backup_route_type;
 	snd_flags_t fwd_snd_flags_bak;
 	snd_flags_t rpl_snd_flags_bak;
@@ -219,6 +230,15 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 	path_bak.s=0;
 	path_bak.len=0;
 	free_path=0;
+	instance_bak.s=0;
+	instance_bak.len=0;
+	free_instance=0;
+	ruid_bak.s=0;
+	ruid_bak.len=0;
+	free_ruid=0;
+	ua_bak.s=0;
+	ua_bak.len=0;
+	free_ua=0;
 	dst=&t->uac[branch].request.dst;
 
 	/* ... we calculate branch ... */	
@@ -259,6 +279,9 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 	parsed_uri_bak=i_req->parsed_uri;
 	parsed_uri_ok_bak=i_req->parsed_uri_ok;
 	path_bak=i_req->path_vec;
+	instance_bak=i_req->instance;
+	ruid_bak=i_req->ruid;
+	ua_bak=i_req->location_ua;
 	
 	if (unlikely(branch_route || has_tran_tmcbs(t, TMCB_REQUEST_FWDED))){
 		/* dup uris, path a.s.o. if we have a branch route or callback */
@@ -292,6 +315,53 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 			}
 			free_path=1;
 		}
+		/* update instance */
+		/* if instance points to msg instance, it needs to be "fixed" so that we 
+		   can change/update msg->instance */
+		if (instance==&i_req->instance)
+			instance=&instance_bak;
+		/* zero it first so that set_instance will work */
+		i_req->instance.s=0;
+		i_req->instance.len=0;
+		if (unlikely(instance)){
+			if (unlikely(set_instance(i_req, instance)<0)){
+				ret=E_OUT_OF_MEM;
+				goto error03;
+			}
+			free_instance=1;
+		}
+
+		/* update ruid */
+		/* if ruid points to msg ruid, it needs to be "fixed" so that we 
+		   can change/update msg->ruid */
+		if (ruid==&i_req->ruid)
+			ruid=&ruid_bak;
+		/* zero it first so that set_ruid will work */
+		i_req->ruid.s=0;
+		i_req->ruid.len=0;
+		if (unlikely(ruid)){
+			if (unlikely(set_ruid(i_req, ruid)<0)){
+				ret=E_OUT_OF_MEM;
+				goto error03;
+			}
+			free_ruid=1;
+		}
+
+		/* update location_ua */
+		/* if location_ua points to msg location_ua, it needs to be "fixed" so that we 
+		   can change/update msg->location_ua */
+		if (location_ua==&i_req->location_ua)
+			location_ua=&ua_bak;
+		/* zero it first so that set_ua will work */
+		i_req->location_ua.s=0;
+		i_req->location_ua.len=0;
+		if (unlikely(location_ua)){
+			if (unlikely(set_ua(i_req, location_ua)<0)){
+				ret=E_OUT_OF_MEM;
+				goto error03;
+			}
+			free_ua=1;
+		}
 	
 		/* backup dst uri  & zero it*/
 		dst_uri_bak=i_req->dst_uri;
@@ -317,7 +387,7 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 			/* run branch_route actions if provided */
 			backup_route_type = get_route_type();
 			set_route_type(BRANCH_ROUTE);
-			tm_ctx_set_branch_index(branch+1);
+			tm_ctx_set_branch_index(branch);
 			/* no need to backup/set avp lists: the on_branch route is run
 			   only in the main route context (e.g. t_relay() in the main
 			   route) or in the failure route context (e.g. append_branch &
@@ -351,14 +421,14 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 				/* if DROP was called in cfg, don't forward, jump to end */
 				if (unlikely(ctx.run_flags&DROP_R_F))
 				{
-					tm_ctx_set_branch_index(0);
+					tm_ctx_set_branch_index(T_BR_UNDEFINED);
 					set_route_type(backup_route_type);
 					/* triggered by drop in CFG */
 					ret=E_CFG;
 					goto error03;
 				}
 			}
-			tm_ctx_set_branch_index(0);
+			tm_ctx_set_branch_index(T_BR_UNDEFINED);
 			set_route_type(backup_route_type);
 		}
 
@@ -397,6 +467,27 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 			i_req->path_vec.s=0;
 			i_req->path_vec.len=0;
 		}
+		if (unlikely(instance && (i_req->instance.s!=instance->s ||
+							  i_req->instance.len!=instance->len))){
+			i_req->instance=*instance;
+		}else if (unlikely(instance==0 && i_req->instance.len!=0)){
+			i_req->instance.s=0;
+			i_req->instance.len=0;
+		}
+		if (unlikely(ruid && (i_req->ruid.s!=ruid->s ||
+							  i_req->ruid.len!=ruid->len))){
+			i_req->ruid=*ruid;
+		}else if (unlikely(ruid==0 && i_req->ruid.len!=0)){
+			i_req->ruid.s=0;
+			i_req->ruid.len=0;
+		}
+		if (unlikely(location_ua && (i_req->location_ua.s!=location_ua->s ||
+							  i_req->location_ua.len!=location_ua->len))){
+			i_req->location_ua=*location_ua;
+		}else if (unlikely(location_ua==0 && i_req->location_ua.len!=0)){
+			i_req->location_ua.s=0;
+			i_req->location_ua.len=0;
+		}
 	}
 	
 	if (likely(next_hop!=0 || (flags & UAC_DNS_FAILOVER_F))){
@@ -416,7 +507,12 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 	} /* else next_hop==0 =>
 		no dst_uri / empty dst_uri and initial next_hop==0 =>
 		dst is pre-filled with a valid dst => use the pre-filled dst */
-	
+
+	/* Set on_reply and on_negative handlers for this branch to the handlers in the transaction */
+	t->uac[branch].on_reply = t->on_reply;
+	t->uac[branch].on_failure = t->on_failure;
+	t->uac[branch].on_branch_failure = t->on_branch_failure;
+
 	/* check if send_sock is ok */
 	if (t->uac[branch].request.dst.send_sock==0) {
 		LOG(L_ERR, "ERROR: can't fwd to af %d, proto %d "
@@ -428,7 +524,7 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 	/* ... and build it now */
 	shbuf=build_req_buf_from_sip_req( i_req, &len, dst, BUILD_IN_SHM);
 	if (!shbuf) {
-		LOG(L_ERR, "ERROR: print_uac_request: no shm mem\n"); 
+		LM_ERR("could not build request\n"); 
 		ret=E_OUT_OF_MEM;
 		goto error01;
 	}
@@ -459,6 +555,60 @@ static int prepare_new_uac( struct cell *t, struct sip_msg *i_req,
 		t->uac[branch].path.s[i_req->path_vec.len]=0;
 		memcpy( t->uac[branch].path.s, i_req->path_vec.s, i_req->path_vec.len);
 	}
+	if (unlikely(i_req->instance.s && i_req->instance.len)){
+		t->uac[branch].instance.s=shm_malloc(i_req->instance.len+1);
+		if (unlikely(t->uac[branch].instance.s==0)) {
+			shm_free(shbuf);
+			t->uac[branch].request.buffer=0;
+			t->uac[branch].request.buffer_len=0;
+			t->uac[branch].uri.s=0;
+			t->uac[branch].uri.len=0;
+			ret=E_OUT_OF_MEM;
+			goto error01;
+		}
+		t->uac[branch].instance.len=i_req->instance.len;
+		t->uac[branch].instance.s[i_req->instance.len]=0;
+		memcpy( t->uac[branch].instance.s, i_req->instance.s, i_req->instance.len);
+	}
+	if (unlikely(i_req->ruid.s && i_req->ruid.len)){
+		t->uac[branch].ruid.s=shm_malloc(i_req->ruid.len+1);
+		if (unlikely(t->uac[branch].ruid.s==0)) {
+			shm_free(shbuf);
+			t->uac[branch].request.buffer=0;
+			t->uac[branch].request.buffer_len=0;
+			t->uac[branch].uri.s=0;
+			t->uac[branch].uri.len=0;
+			ret=E_OUT_OF_MEM;
+			goto error01;
+		}
+		t->uac[branch].ruid.len=i_req->ruid.len;
+		t->uac[branch].ruid.s[i_req->ruid.len]=0;
+		memcpy( t->uac[branch].ruid.s, i_req->ruid.s, i_req->ruid.len);
+	}
+	if (unlikely(i_req->location_ua.s && i_req->location_ua.len)){
+		t->uac[branch].location_ua.s=shm_malloc(i_req->location_ua.len+1);
+		if (unlikely(t->uac[branch].location_ua.s==0)) {
+			shm_free(shbuf);
+			t->uac[branch].request.buffer=0;
+			t->uac[branch].request.buffer_len=0;
+			t->uac[branch].uri.s=0;
+			t->uac[branch].uri.len=0;
+			ret=E_OUT_OF_MEM;
+			goto error01;
+		}
+		t->uac[branch].location_ua.len=i_req->location_ua.len;
+		t->uac[branch].location_ua.s[i_req->location_ua.len]=0;
+		memcpy( t->uac[branch].location_ua.s, i_req->location_ua.s, i_req->location_ua.len);
+	}
+
+#ifdef TM_UAC_FLAGS
+	len = count_applied_lumps(i_req->add_rm, HDR_RECORDROUTE_T);
+	if(len==1)
+		t->uac[branch].flags = TM_UAC_FLAG_RR;
+	else if(len==2)
+		t->uac[branch].flags = TM_UAC_FLAG_RR|TM_UAC_FLAG_R2;
+#endif
+
 	ret=0;
 
 error01:
@@ -470,6 +620,15 @@ error03:
 	if (unlikely(free_path)){
 		reset_path_vector(i_req);
 	}
+	if (unlikely(free_instance)){
+		reset_instance(i_req);
+	}
+	if (unlikely(free_ruid)){
+		reset_ruid(i_req);
+	}
+	if (unlikely(free_ua)){
+		reset_ua(i_req);
+	}
 	if (dst_uri_backed_up){
 		reset_dst_uri(i_req); /* free dst_uri */
 		i_req->dst_uri=dst_uri_bak;
@@ -479,6 +638,9 @@ error03:
 	i_req->parsed_uri=parsed_uri_bak;
 	i_req->parsed_uri_ok=parsed_uri_ok_bak;
 	i_req->path_vec=path_bak;
+	i_req->instance=instance_bak;
+	i_req->ruid=ruid_bak;
+	i_req->location_ua=ua_bak;
 	
 	/* Delete the duplicated lump lists, this will also delete
 	 * all lumps created here, such as lumps created in per-branch
@@ -603,6 +765,11 @@ int add_blind_uac( /*struct cell *t*/ )
 	membar_write(); /* to allow lockless prepare_to_cancel() we want to be sure
 					   all the writes finished before updating branch number*/
 	t->nr_of_outgoings=(branch+1);
+	t->async_backup.blind_uac = branch; /* whenever we create a blind UAC, lets save the current branch
+					 * this is used in async tm processing specifically to be able to route replies
+					 * that were possibly in response to a request forwarded on this blind UAC......
+					 * we still want replies to be processed as if it were a normal UAC */
+	
 	/* start FR timer -- protocol set by default to PROTO_NONE,
        which means retransmission timer will not be started
     */
@@ -639,17 +806,15 @@ int add_blind_uac( /*struct cell *t*/ )
  *                    @see prepare_new_uac().
  *  @returns branch id (>=0) or error (<0)
 */
-static int add_uac( struct cell *t, struct sip_msg *request, str *uri,
+int add_uac( struct cell *t, struct sip_msg *request, str *uri,
 					str* next_hop, str* path, struct proxy_l *proxy,
 					struct socket_info* fsocket, snd_flags_t snd_flags,
-					int proto, int flags)
+					int proto, int flags, str *instance, str *ruid,
+					str *location_ua)
 {
 
 	int ret;
 	unsigned short branch;
-#ifdef TM_UAC_FLAGS
-	unsigned int len;
-#endif /* TM_UAC_FLAGS */
 
 	branch=t->nr_of_outgoings;
 	if (branch==MAX_BRANCHES) {
@@ -687,17 +852,11 @@ static int add_uac( struct cell *t, struct sip_msg *request, str *uri,
 	/* now message printing starts ... */
 	if (unlikely( (ret=prepare_new_uac(t, request, branch, uri, path,
 										next_hop, fsocket, snd_flags,
-										proto, flags)) < 0)){
+										proto, flags, instance, ruid,
+										location_ua)) < 0)){
 		ser_error=ret;
 		goto error01;
 	}
-#ifdef TM_UAC_FLAGS
-	len = count_applied_lumps(request->add_rm, HDR_RECORDROUTE_T);
-	if(len==1)
-		t->uac[branch].flags = TM_UAC_FLAG_RR;
-	else if(len==2)
-		t->uac[branch].flags = TM_UAC_FLAG_RR|TM_UAC_FLAG_R2;
-#endif
 	getbflagsval(0, &t->uac[branch].branch_flags);
 	membar_write(); /* to allow lockless ops (e.g. prepare_to_cancel()) we want
 					   to be sure everything above is fully written before
@@ -727,7 +886,9 @@ static int add_uac_from_buf( struct cell *t, struct sip_msg *request,
 								struct socket_info* fsocket,
 								snd_flags_t send_flags,
 								int proto,
-								char *buf, short buf_len)
+								char *buf, short buf_len,
+								str *instance, str *ruid,
+								str *location_ua)
 {
 
 	int ret;
@@ -798,6 +959,59 @@ static int add_uac_from_buf( struct cell *t, struct sip_msg *request,
 		t->uac[branch].path.s[path->len]=0;
 		memcpy( t->uac[branch].path.s, path->s, path->len);
 	}
+	/* copy the instance */
+	if (unlikely(instance && instance->s)){
+		t->uac[branch].instance.s=shm_malloc(instance->len+1);
+		if (unlikely(t->uac[branch].instance.s==0)) {
+			shm_free(shbuf);
+			t->uac[branch].request.buffer=0;
+			t->uac[branch].request.buffer_len=0;
+			t->uac[branch].uri.s=0;
+			t->uac[branch].uri.len=0;
+			ret=ser_error=E_OUT_OF_MEM;
+			goto error;
+		}
+		t->uac[branch].instance.len=instance->len;
+		t->uac[branch].instance.s[instance->len]=0;
+		memcpy( t->uac[branch].instance.s, instance->s, instance->len);
+	}
+	/* copy the ruid */
+	if (unlikely(ruid && ruid->s)){
+		t->uac[branch].ruid.s=shm_malloc(ruid->len+1);
+		if (unlikely(t->uac[branch].ruid.s==0)) {
+			shm_free(shbuf);
+			t->uac[branch].request.buffer=0;
+			t->uac[branch].request.buffer_len=0;
+			t->uac[branch].uri.s=0;
+			t->uac[branch].uri.len=0;
+			ret=ser_error=E_OUT_OF_MEM;
+			goto error;
+		}
+		t->uac[branch].ruid.len=ruid->len;
+		t->uac[branch].ruid.s[ruid->len]=0;
+		memcpy( t->uac[branch].ruid.s, ruid->s, ruid->len);
+	}
+	/* copy the location_ua */
+	if (unlikely(location_ua && location_ua->s)){
+		t->uac[branch].location_ua.s=shm_malloc(location_ua->len+1);
+		if (unlikely(t->uac[branch].location_ua.s==0)) {
+			shm_free(shbuf);
+			t->uac[branch].request.buffer=0;
+			t->uac[branch].request.buffer_len=0;
+			t->uac[branch].uri.s=0;
+			t->uac[branch].uri.len=0;
+			ret=ser_error=E_OUT_OF_MEM;
+			goto error;
+		}
+		t->uac[branch].location_ua.len=location_ua->len;
+		t->uac[branch].location_ua.s[location_ua->len]=0;
+		memcpy( t->uac[branch].location_ua.s, location_ua->s, location_ua->len);
+	}
+
+	t->uac[branch].on_reply = t->on_reply;
+	t->uac[branch].on_failure = t->on_failure;
+	t->uac[branch].on_branch_failure = t->on_branch_failure;
+
 	membar_write(); /* to allow lockless ops (e.g. prepare_to_cancel()) we want
 					   to be sure everything above is fully written before
 					   updating branches no. */
@@ -865,12 +1079,16 @@ int add_uac_dns_fallback(struct cell *t, struct sip_msg* msg,
 							&old_uac->path,
 							 (old_uac->request.dst.send_flags.f &
 								SND_F_FORCE_SOCKET)?
-									old_uac->request.dst.send_sock:0,
+									old_uac->request.dst.send_sock:
+									((tm_dns_reuse_rcv_socket)
+											?msg->rcv.bind_address:0),
 							old_uac->request.dst.send_flags,
 							old_uac->request.dst.proto,
 							old_uac->request.buffer,
-							old_uac->request.buffer_len);
-			}else
+							old_uac->request.buffer_len,
+							&old_uac->instance, &old_uac->ruid,
+							&old_uac->location_ua);
+			} else {
 				/* add_uac will use dns_h => next_hop will be ignored.
 				 * Unfortunately we can't reuse the old buffer, the branch id
 				 *  must be changed and the send_socket might be different =>
@@ -878,9 +1096,14 @@ int add_uac_dns_fallback(struct cell *t, struct sip_msg* msg,
 				ret=add_uac(t,  msg, &old_uac->uri, 0, &old_uac->path, 0,
 							 (old_uac->request.dst.send_flags.f &
 								SND_F_FORCE_SOCKET)?
-									old_uac->request.dst.send_sock:0,
+									old_uac->request.dst.send_sock:
+									((tm_dns_reuse_rcv_socket)
+											?msg->rcv.bind_address:0),
 							old_uac->request.dst.send_flags,
-							old_uac->request.dst.proto, UAC_DNS_FAILOVER_F);
+							old_uac->request.dst.proto, UAC_DNS_FAILOVER_F,
+							&old_uac->instance, &old_uac->ruid,
+							&old_uac->location_ua);
+			}
 
 			if (ret<0){
 				/* failed, delete the copied dns_h */
@@ -955,7 +1178,8 @@ int e2e_cancel_branch( struct sip_msg *cancel_msg, struct cell *t_cancel,
 		if (unlikely((ret=prepare_new_uac( t_cancel, cancel_msg, branch,
 									&t_invite->uac[branch].uri,
 									&t_invite->uac[branch].path,
-									0, 0, snd_flags, PROTO_NONE, 0)) <0)){
+									0, 0, snd_flags, PROTO_NONE, 0,
+									NULL, NULL, NULL)) <0)){
 			ser_error=ret;
 			goto error;
 		}
@@ -1379,9 +1603,25 @@ int t_send_branch( struct cell *t, int branch, struct sip_msg* p_msg ,
 			}
 		}
 #endif
-		LOG(L_ERR, "ERROR: t_send_branch: sending request on branch %d "
+		uac->icode = 908; /* internal code set to delivery failure */
+		LOG(L_WARN, "ERROR: t_send_branch: sending request on branch %d "
 				"failed\n", branch);
 		if (proxy) { proxy->errors++; proxy->ok=0; }
+		if(tm_failure_exec_mode==1) {
+			LM_DBG("putting branch %d on hold \n", branch);
+			/* put on retransmission timer,
+			 * but set proto to NONE, so actually it is not trying to resend */
+			uac->request.dst.proto = PROTO_NONE;
+			/* reset last_received, 408 reply is faked by timer */
+			uac->last_received=0;
+			/* add to retransmission timer */
+			if (start_retr( &uac->request )!=0){
+				LM_CRIT("retransmission already started for %p\n",
+					&uac->request);
+				return -2;
+			}
+			return branch;
+		}
 		return -2;
 	} else {
 		if (unlikely(has_tran_tmcbs(t, TMCB_REQUEST_SENT)))
@@ -1414,7 +1654,7 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 	int success_branch;
 	int try_new;
 	int lock_replies;
-	str dst_uri, path;
+	str dst_uri, path, instance, ruid, location_ua;
 	struct socket_info* si;
 	flag_t backup_bflags = 0;
 	flag_t bflags = 0;
@@ -1423,11 +1663,10 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 	/* make -Wall happy */
 	current_uri.s=0;
 
-	if (t->flags & T_CANCELED){
-		DBG("t_forward_non_ack: no forwarding on a canceled transaction\n");
-		ser_error=E_CANCELED;
-		return -1;
-	}
+	getbflagsval(0, &backup_bflags);
+
+	if (t->flags & T_CANCELED) goto canceled;
+
 	if (p_msg->REQ_METHOD==METHOD_CANCEL) { 
 		t_invite=t_lookupOriginalT(  p_msg );
 		if (t_invite!=T_NULL_CELL) {
@@ -1440,8 +1679,6 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 		}
 	}
 
-	getbflagsval(0, &backup_bflags);
-
 	/* if no more specific error code is known, use this */
 	lowest_ret=E_UNSPEC;
 	/* branches added */
@@ -1452,6 +1689,10 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 	if (t->on_branch) {
 		/* tell add_uac that it should run branch route actions */
 		branch_route = t->on_branch;
+		/* save the branch route so that it
+		 * can be used for adding branches later
+		 */
+		t->on_branch_delayed = t->on_branch;
 		/* reset the flag before running the actions (so that it
 		 * could be set again in branch_route if needed
 		 */
@@ -1480,7 +1721,10 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 		branch_ret=add_uac( t, p_msg, GET_RURI(p_msg), GET_NEXT_HOP(p_msg),
 							&p_msg->path_vec, proxy, p_msg->force_send_socket,
 							p_msg->fwd_send_flags, proto,
-							(p_msg->dst_uri.len)?0:UAC_SKIP_BR_DST_F);
+							(p_msg->dst_uri.len)?0:UAC_SKIP_BR_DST_F, &p_msg->instance,
+							&p_msg->ruid, &p_msg->location_ua);
+		/* test if cancel was received meanwhile */
+		if (t->flags & T_CANCELED) goto canceled;
 		if (branch_ret>=0) 
 			added_branches |= 1<<branch_ret;
 		else
@@ -1489,14 +1733,17 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 
 	init_branch_iterator();
 	while((current_uri.s=next_branch( &current_uri.len, &q, &dst_uri, &path,
-										&bflags, &si))) {
+										&bflags, &si, &ruid, &instance, &location_ua))) {
 		try_new++;
 		setbflagsval(0, bflags);
 
 		branch_ret=add_uac( t, p_msg, &current_uri,
 							(dst_uri.len) ? (&dst_uri) : &current_uri,
 							&path, proxy, si, p_msg->fwd_send_flags,
-							proto, (dst_uri.len)?0:UAC_SKIP_BR_DST_F);
+							proto, (dst_uri.len)?0:UAC_SKIP_BR_DST_F, &instance,
+							&ruid, &location_ua);
+		/* test if cancel was received meanwhile */
+		if (t->flags & T_CANCELED) goto canceled;
 		/* pick some of the errors in case things go wrong;
 		   note that picking lowest error is just as good as
 		   any other algorithm which picks any other negative
@@ -1570,6 +1817,17 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 	ser_error=0; /* clear branch send errors, we have overall success */
 	set_kr(REQ_FWDED);
 	return 1;
+
+canceled:
+	DBG("t_forward_non_ack: no forwarding on a canceled transaction\n");
+	/* reset processed branches */
+	clear_branches();
+	/* restore backup flags from initial env */
+	setbflagsval(0, backup_bflags);
+	/* update message flags, if changed in branch route */
+	t->uas.request->flags = p_msg->flags;
+	ser_error=E_CANCELED;
+	return -1;
 }
 
 
